"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.IconService = void 0;
const tslib_1 = require("tslib");
const di_1 = require("@opensumi/di");
const ide_core_browser_1 = require("@opensumi/ide-core-browser");
const static_resource_1 = require("@opensumi/ide-core-browser/lib/static-resource");
const common_1 = require("../common");
const icon_theme_store_1 = require("./icon-theme-store");
require("./icon.less");
let IconService = class IconService extends ide_core_browser_1.WithEventBus {
    getPath(basePath, relativePath) {
        if (relativePath.startsWith('./')) {
            const uri = new ide_core_browser_1.URI(basePath).resolve(relativePath.replace(/^\.\//, ''));
            return uri.scheme ? uri : ide_core_browser_1.URI.file(uri.toString());
        }
        else if (/^http(s)?/.test(relativePath)) {
            return new ide_core_browser_1.URI(relativePath);
        }
        else if (basePath) {
            const uri = new ide_core_browser_1.URI(basePath).resolve(relativePath);
            return uri.scheme ? uri : ide_core_browser_1.URI.file(uri.toString());
        }
        else if (/^file:\/\//.test(relativePath)) {
            return new ide_core_browser_1.URI(relativePath);
        }
        else {
            return ide_core_browser_1.URI.file(relativePath);
        }
    }
    constructor() {
        super();
        this.iconThemeLoaded = new ide_core_browser_1.Deferred();
        this.themeChangeEmitter = new ide_core_browser_1.Emitter();
        this.onThemeChange = this.themeChangeEmitter.event;
        this.iconThemes = new Map();
        this.iconContributionRegistry = new Map();
        this.iconMap = new Map();
        // eg. $(codicon/sync~spin)
        this._regexFromString = /^\$\(([a-z.]+\/)?([a-z-]+)(~[a-z]+)?\)$/i;
        this.styleSheetCollection = '';
        this.appendStyleCounter = 0;
        this.listen();
    }
    listen() {
        this.preferenceService.onPreferenceChanged(async (e) => {
            if (e.preferenceName === "general.icon" /* GeneralSettingsId.Icon */ && this.iconContributionRegistry.has(e.newValue)) {
                await this.applyTheme(this.preferenceService.get("general.icon" /* GeneralSettingsId.Icon */));
            }
        });
    }
    doAppend(targetElement) {
        if (targetElement) {
            const textContent = (targetElement === null || targetElement === void 0 ? void 0 : targetElement.textContent) + this.styleSheetCollection;
            targetElement.textContent = textContent;
        }
        this.styleSheetCollection = '';
        this.appendStylesTimer = undefined;
        this.appendStyleCounter = 0;
        clearTimeout(this.appendStylesTimer);
    }
    appendStyleSheet(styleSheet, fromExtension = false) {
        let iconStyleNode = document.getElementById('plugin-icons');
        if (!iconStyleNode) {
            iconStyleNode = document.createElement('style');
            iconStyleNode.id = 'plugin-icons';
            document.getElementsByTagName('head')[0].appendChild(iconStyleNode);
        }
        // 非插件进程注册的 icon 正常 append
        if (!fromExtension) {
            iconStyleNode.append(styleSheet);
            return;
        }
        // 针对插件进程注册的 icon 进行 append 分段处理
        // 避免因为过多 icon 导致页面卡顿
        // 例如 GitLens 插件会注册超过 800 个 icon
        this.styleSheetCollection += '\r\n' + styleSheet;
        this.appendStyleCounter += 1;
        // 超过 100 个样式
        if (this.appendStyleCounter >= 150 && this.appendStylesTimer) {
            clearTimeout(this.appendStylesTimer);
            this.appendExtensionIconStyle();
        }
        if (!this.appendStylesTimer) {
            // 超过 100 毫秒
            this.appendStylesTimer = window.setTimeout(() => {
                this.appendExtensionIconStyle();
            }, 100);
        }
    }
    appendExtensionIconStyle(styleNode) {
        if (styleNode) {
            this.doAppend(styleNode);
        }
        else {
            const randomClass = this.getRandomIconClass('extension-');
            const styleNode = document.createElement('style');
            styleNode.className = randomClass;
            this.doAppend(styleNode);
            document.getElementsByTagName('head')[0].appendChild(styleNode);
        }
    }
    getRandomIconClass(prefix = '') {
        return `${prefix}icon-${(0, ide_core_browser_1.randomString)(6)}`;
    }
    getMaskStyleSheet(iconUrl, className, baseTheme) {
        const cssRule = `${baseTheme || ''} .${className} {-webkit-mask: url("${iconUrl}") no-repeat 50% 50%;}`;
        return cssRule;
    }
    getMaskStyleSheetWithStaticService(path, className, baseTheme) {
        const iconUrl = this.staticResourceService.resolveStaticResource(path).toString();
        return this.getMaskStyleSheet(iconUrl, className, baseTheme);
    }
    getBackgroundStyleSheet(iconUrl, className, baseTheme) {
        const cssRule = `${baseTheme || ''} .${className} {background: url("${iconUrl}") no-repeat 0 0;background-size: contain;}`;
        return cssRule;
    }
    getBackgroundStyleSheetWithStaticService(path, className, baseTheme) {
        const iconUrl = this.staticResourceService.resolveStaticResource(path).toString();
        return this.getBackgroundStyleSheet(iconUrl, className, baseTheme);
    }
    fromString(str) {
        if (typeof str !== 'string') {
            return undefined;
        }
        const matched = str.match(this._regexFromString);
        if (!matched) {
            return undefined;
        }
        const [, owner, name, modifier] = matched;
        const iconOwner = owner ? owner.slice(0, -1) : ide_core_browser_1.CODICON_OWNER;
        let className = `${iconOwner} ${iconOwner}-${name}`;
        if (modifier) {
            className += ` ${modifier.slice(1)}`;
        }
        return className;
    }
    encodeBase64Path(iconPath) {
        // 由于进行 Background 样式拼接时采用的 `url("${iconPath}")` 结构
        // 故这里需要对 iconPath 中的 `"` 及常见字符进行转义处理
        return iconPath.replace(/"/g, '\\"').replace(/\?|#/g, (m) => encodeURIComponent(m));
    }
    fromIcon(basePath = '', icon, type = common_1.IconType.Mask, shape = common_1.IconShape.Square, fromExtension = false) {
        if (!icon) {
            return;
        }
        const iconPath = typeof icon === 'string' ? icon : icon.dark;
        const iconId = `${basePath}-${iconPath}-${type}-${shape}`;
        if (this.iconMap.get(iconId)) {
            return this.iconMap.get(iconId);
        }
        const randomClass = this.getRandomIconClass();
        if (typeof icon === 'string') {
            /**
             * 处理 data:image 格式，/^data:image\//
             * 如 data:image/svg+xml or data:image/gif;base64,
             * 此时无需 static service
             */
            if (type === common_1.IconType.Base64) {
                this.appendStyleSheet(this.getBackgroundStyleSheet(this.encodeBase64Path(icon), randomClass), fromExtension);
            }
            else {
                const targetPath = this.getPath(basePath, icon);
                if (type === common_1.IconType.Mask) {
                    this.appendStyleSheet(this.getMaskStyleSheetWithStaticService(targetPath, randomClass), fromExtension);
                }
                else {
                    this.appendStyleSheet(this.getBackgroundStyleSheetWithStaticService(targetPath, randomClass), fromExtension);
                }
            }
        }
        else {
            // eslint-disable-next-line guard-for-in
            for (const themeType in icon) {
                const themeSelector = (0, common_1.getThemeTypeSelector)(themeType);
                const iconPath = icon[themeType];
                const targetPath = this.getPath(basePath, iconPath);
                if (type === common_1.IconType.Base64) {
                    /**
                     * 处理 data:image 格式，/^data:image\//
                     * 如 data:image/svg+xml or data:image/gif;base64,
                     * 此时无需 static service
                     */
                    this.appendStyleSheet(this.getBackgroundStyleSheet(this.encodeBase64Path(iconPath), randomClass, `.${themeSelector}`), fromExtension);
                }
                else if (type === common_1.IconType.Mask) {
                    this.appendStyleSheet(this.getMaskStyleSheetWithStaticService(targetPath, randomClass, `.${themeSelector}`), fromExtension);
                }
                else {
                    this.appendStyleSheet(this.getBackgroundStyleSheetWithStaticService(targetPath, randomClass, `.${themeSelector}`), fromExtension);
                }
            }
        }
        const targetIconClass = [
            'kaitian-icon',
            randomClass,
            {
                [common_1.IconType.Background]: 'background-mode',
                [common_1.IconType.Base64]: 'background-mode',
                [common_1.IconType.Mask]: 'mask-mode',
            }[type],
            shape === common_1.IconShape.Circle ? 'circle' : '',
        ].join(' ');
        this.iconMap.set(iconId, targetIconClass);
        return targetIconClass;
    }
    async onDidExtensionContributes() {
        await this.updateIconThemes();
        this.iconThemeLoaded.resolve();
    }
    get preferenceThemeId() {
        return this.preferenceService.get("general.icon" /* GeneralSettingsId.Icon */);
    }
    async updateIconThemes() {
        const themeMap = this.getAvailableThemeInfos().reduce((pre, cur) => {
            if (!pre.has(cur.themeId)) {
                pre.set(cur.themeId, cur.name);
            }
            return pre;
        }, new Map());
        this.preferenceSettings.setEnumLabels("general.icon" /* GeneralSettingsId.Icon */, Object.fromEntries(themeMap.entries()));
        // 当前没有主题，或没有缓存的主题时，将第一个注册主题设置为当前主题
        if (!this.currentTheme) {
            if (!this.preferenceThemeId || !themeMap.has(this.preferenceThemeId)) {
                const themeId = Array.from(themeMap.keys())[0];
                if (themeId) {
                    await this.applyTheme(themeId);
                }
            }
        }
    }
    registerIconThemes(iconContributions, basePath) {
        for (const contribution of iconContributions) {
            const themeId = (0, common_1.getThemeId)(contribution);
            this.iconContributionRegistry.set(themeId, { contribution, basePath });
            if (this.preferenceThemeId && this.preferenceThemeId === themeId) {
                this.applyTheme(this.preferenceThemeId);
            }
        }
        const currentSchemas = this.preferenceSchemaProvider.getPreferenceProperty("general.icon" /* GeneralSettingsId.Icon */);
        if (currentSchemas) {
            delete currentSchemas.scope;
        }
        this.preferenceSchemaProvider.setSchema({
            properties: {
                ["general.icon" /* GeneralSettingsId.Icon */]: Object.assign(Object.assign({}, currentSchemas), { enum: this.getAvailableThemeInfos().map((info) => info.themeId) }),
            },
        }, true);
        this.updateIconThemes();
    }
    getAvailableThemeInfos() {
        const themeInfos = [];
        for (const { contribution } of this.iconContributionRegistry.values()) {
            const { label, id, extensionId } = contribution;
            themeInfos.push({
                themeId: id || (0, common_1.getThemeId)(contribution),
                name: label,
                extensionId,
            });
        }
        return themeInfos;
    }
    async getIconTheme(themeId) {
        let theme = this.iconThemes.get(themeId);
        if (theme) {
            return theme;
        }
        const extContribution = this.iconContributionRegistry.get(themeId);
        if (extContribution) {
            theme = await this.iconThemeStore.getIconTheme(extContribution.contribution, extContribution.basePath);
            return theme;
        }
        return;
    }
    async applyTheme(themeId) {
        this.toggleIconVisible(true);
        if (this.currentTheme && this.currentThemeId === themeId) {
            return;
        }
        /**
         * 这里 `applyTheme` 默认应该按照最后一个应用的主题进行加载
         * 但由于 `getIconTheme` 存在时序问题，例如：
         * 主题 A，E，分别由插件 A，E 贡献
         * 这里先调用 applyTheme(E), 再调用 applyTheme(A)
         * 旧的逻辑由于插件 A ... E 的加载顺序问题，会存在 A 比 E 快加载的情况导致最终应用了错误的主题
         *
         * 故这里增加额外判断，保障最后一个加载的主题应用
         */
        this.latestApplyTheme = themeId;
        const iconThemeData = await this.getIconTheme(themeId);
        if (this.latestApplyTheme !== themeId) {
            return;
        }
        if (!iconThemeData) {
            this.logger.warn('Target IconTheme extension not detected, use built-in icons.');
            document.getElementsByTagName('body')[0].classList.add('default-file-icons');
            this.iconThemeLoaded.resolve();
            return;
        }
        this.currentThemeId = themeId;
        document.getElementsByTagName('body')[0].classList.remove('default-file-icons');
        this.currentTheme = iconThemeData;
        const { styleSheetContent } = iconThemeData;
        let styleNode = document.getElementById('icon-style');
        if (styleNode) {
            styleNode.innerHTML = styleSheetContent;
        }
        else {
            styleNode = document.createElement('style');
            styleNode.id = 'icon-style';
            styleNode.innerHTML = styleSheetContent;
            document.getElementsByTagName('head')[0].appendChild(styleNode);
        }
        this.themeChangeEmitter.fire(this.currentTheme);
        if (!this.preferenceThemeId) {
            this.iconThemeLoaded.resolve();
        }
    }
    toggleIconVisible(show) {
        const rootNode = document.getElementsByTagName('body')[0];
        if (show === undefined) {
            rootNode.classList.toggle('show-file-icons');
        }
        else if (show === true) {
            rootNode.classList.add('show-file-icons');
        }
        else {
            rootNode.classList.remove('show-file-icons');
        }
    }
};
tslib_1.__decorate([
    (0, di_1.Autowired)(),
    tslib_1.__metadata("design:type", static_resource_1.StaticResourceService)
], IconService.prototype, "staticResourceService", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(),
    tslib_1.__metadata("design:type", icon_theme_store_1.IconThemeStore)
], IconService.prototype, "iconThemeStore", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(ide_core_browser_1.PreferenceService),
    tslib_1.__metadata("design:type", Object)
], IconService.prototype, "preferenceService", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(ide_core_browser_1.PreferenceSchemaProvider),
    tslib_1.__metadata("design:type", ide_core_browser_1.PreferenceSchemaProvider)
], IconService.prototype, "preferenceSchemaProvider", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(ide_core_browser_1.IPreferenceSettingsService),
    tslib_1.__metadata("design:type", Object)
], IconService.prototype, "preferenceSettings", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(ide_core_browser_1.ILogger),
    tslib_1.__metadata("design:type", Object)
], IconService.prototype, "logger", void 0);
tslib_1.__decorate([
    (0, ide_core_browser_1.OnEvent)(ide_core_browser_1.ExtensionDidContributes),
    tslib_1.__metadata("design:type", Function),
    tslib_1.__metadata("design:paramtypes", []),
    tslib_1.__metadata("design:returntype", Promise)
], IconService.prototype, "onDidExtensionContributes", null);
IconService = tslib_1.__decorate([
    (0, di_1.Injectable)(),
    tslib_1.__metadata("design:paramtypes", [])
], IconService);
exports.IconService = IconService;
//# sourceMappingURL=icon.service.js.map