"use strict";
var ProductIconThemeData_1;
Object.defineProperty(exports, "__esModule", { value: true });
exports.ProductIconThemeData = exports.fontFormatRegex = exports.fontSizeRegex = exports.fontWeightRegex = exports.fontStyleRegex = exports.fontIdRegex = void 0;
const tslib_1 = require("tslib");
const parser = tslib_1.__importStar(require("jsonc-parser"));
const di_1 = require("@opensumi/di");
const ide_core_browser_1 = require("@opensumi/ide-core-browser");
const ide_file_service_1 = require("@opensumi/ide-file-service");
const icon_registry_1 = require("../common/icon-registry");
exports.fontIdRegex = '^([\\w_-]+)$';
exports.fontStyleRegex = '^(normal|italic|(oblique[ \\w\\s-]+))$';
exports.fontWeightRegex = '^(normal|bold|lighter|bolder|(\\d{0-1000}))$';
exports.fontSizeRegex = '^([\\w .%_-]+)$';
exports.fontFormatRegex = '^woff|woff2|truetype|opentype|embedded-opentype|svg$';
let ProductIconThemeData = ProductIconThemeData_1 = class ProductIconThemeData {
    constructor(id, label, settingsId) {
        this.iconThemeDocument = { iconDefinitions: new Map() };
        this.id = id;
        this.label = label;
        this.settingsId = settingsId;
        this.isLoaded = false;
    }
    async load(location) {
        this.location = location;
        const warnings = [];
        try {
            this.iconThemeDocument = await _loadProductIconThemeDocument(this.fileServiceClient, location, warnings);
        }
        catch (error) {
            // TODO more error handling
            this.logger.error(`parse icon theme failed ${location.codeUri.fsPath}`);
        }
        this.isLoaded = true;
        if (warnings.length) {
            this.logger.warn(warnings.toString());
        }
        return this.styleSheetContent;
    }
    getIcon(iconContribution) {
        return _resolveIconDefinition(iconContribution, this.iconThemeDocument);
    }
    static fromExtensionTheme(iconTheme, iconThemeLocation, extensionData) {
        const id = extensionData.extensionId + '-' + iconTheme.id;
        const label = iconTheme.label;
        const settingsId = iconTheme.id;
        const themeData = new ProductIconThemeData_1(id, label, settingsId);
        themeData.description = iconTheme.description;
        themeData.location = iconThemeLocation;
        themeData.extensionData = extensionData;
        themeData.isLoaded = false;
        return themeData;
    }
};
tslib_1.__decorate([
    (0, di_1.Autowired)(ide_file_service_1.IFileServiceClient),
    tslib_1.__metadata("design:type", Object)
], ProductIconThemeData.prototype, "fileServiceClient", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(),
    tslib_1.__metadata("design:type", ide_core_browser_1.Logger)
], ProductIconThemeData.prototype, "logger", void 0);
ProductIconThemeData = ProductIconThemeData_1 = tslib_1.__decorate([
    (0, di_1.Injectable)({ multiple: true }),
    tslib_1.__metadata("design:paramtypes", [String, String, String])
], ProductIconThemeData);
exports.ProductIconThemeData = ProductIconThemeData;
const iconRegistry = (0, icon_registry_1.getIconRegistry)();
function _resolveIconDefinition(iconContribution, iconThemeDocument) {
    const iconDefinitions = iconThemeDocument.iconDefinitions;
    let definition = iconDefinitions.get(iconContribution.id);
    let defaults = iconContribution.defaults;
    while (!definition && ide_core_browser_1.ThemeIcon.isThemeIcon(defaults)) {
        // sumi icon proxy
        if (defaults.alias) {
            const alias = defaults.alias;
            const definitionProxy = alias.find((iconId) => iconDefinitions.get(iconId));
            return definitionProxy ? iconDefinitions.get(definitionProxy) : undefined;
        }
        // look if an inherited icon has a definition
        const ic = iconRegistry.getIcon(defaults.id);
        if (ic) {
            definition = iconDefinitions.get(ic.id);
            defaults = ic.defaults;
        }
        else {
            return undefined;
        }
    }
    if (definition) {
        return definition;
    }
    if (!ide_core_browser_1.ThemeIcon.isThemeIcon(defaults)) {
        return defaults;
    }
    return undefined;
}
async function _loadProductIconThemeDocument(fileService, location, warnings) {
    var _a;
    const ret = await fileService.readFile(location.toString());
    const parseErrors = [];
    const contentValue = parser.parse(ret.content.toString(), parseErrors);
    if (parseErrors.length > 0) {
        return Promise.reject(new Error('Parse product icon theme failed'));
    }
    else if (typeof contentValue !== 'object') {
        return Promise.reject(new Error('Invalid icon theme format'));
    }
    else if (!contentValue.iconDefinitions || !Array.isArray(contentValue.fonts) || !contentValue.fonts.length) {
        return Promise.reject(new Error('Missing properties: iconDefinitions or fonts'));
    }
    const iconThemeDocumentLocationDirname = ide_core_browser_1.path.dirname(location.toString());
    const sanitizedFonts = new Map();
    for (const font of contentValue.fonts) {
        if ((0, ide_core_browser_1.isString)(font.id) && font.id.match(exports.fontIdRegex)) {
            const fontId = font.id;
            let fontWeight;
            if ((0, ide_core_browser_1.isString)(font.weight) && font.weight.match(exports.fontWeightRegex)) {
                fontWeight = font.weight;
            }
            else {
                warnings.push(`Invalid font weight in font '${font.id}'. Ignoring setting.`);
            }
            let fontStyle;
            if ((0, ide_core_browser_1.isString)(font.style) && font.style.match(exports.fontStyleRegex)) {
                fontStyle = font.style;
            }
            else {
                warnings.push(`Invalid font style in font '${font.id}'. Ignoring setting.`);
            }
            const sanitizedSrc = [];
            if (Array.isArray(font.src)) {
                for (const s of font.src) {
                    if ((0, ide_core_browser_1.isString)(s.path) && (0, ide_core_browser_1.isString)(s.format) && s.format.match(exports.fontFormatRegex)) {
                        const iconFontLocation = ide_core_browser_1.path.join(iconThemeDocumentLocationDirname, s.path);
                        sanitizedSrc.push({ location: ide_core_browser_1.URI.parse(iconFontLocation), format: s.format });
                    }
                    else {
                        warnings.push(`Invalid font source in font '${font.id}'. Ignoring source.`);
                    }
                }
            }
            if (sanitizedSrc.length) {
                sanitizedFonts.set(fontId, { weight: fontWeight, style: fontStyle, src: sanitizedSrc });
            }
            else {
                warnings.push(`No valid font source in font '${font.id}'. Ignoring font definition.`);
            }
        }
        else {
            warnings.push(`Missing or invalid font id '${font.id}'. Skipping font definition.`);
        }
    }
    const iconDefinitions = new Map();
    const primaryFontId = contentValue.fonts[0].id;
    // eslint-disable-next-line guard-for-in
    for (const iconId in contentValue.iconDefinitions) {
        const definition = contentValue.iconDefinitions[iconId];
        if ((0, ide_core_browser_1.isString)(definition.fontCharacter)) {
            const fontId = (_a = definition.fontId) !== null && _a !== void 0 ? _a : primaryFontId;
            const fontDefinition = sanitizedFonts.get(fontId);
            if (fontDefinition) {
                // pi ?
                const font = { id: `pi-${fontId}`, definition: fontDefinition };
                iconDefinitions.set(iconId, { fontCharacter: definition.fontCharacter, font });
            }
            else {
                warnings.push(`Skipping icon definition '${iconId}'. Unknown font.`);
            }
        }
        else {
            warnings.push(`Skipping icon definition '${iconId}'. Unknown fontCharacter.`);
        }
    }
    return { iconDefinitions };
}
//# sourceMappingURL=product-icon-theme-data.js.map