"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SemanticTokenRegistryImpl = void 0;
const tslib_1 = require("tslib");
const di_1 = require("@opensumi/di");
const localize_1 = require("@opensumi/ide-core-common/lib/localize");
const semantic_tokens_registry_1 = require("../common/semantic-tokens-registry");
let SemanticTokenRegistryImpl = class SemanticTokenRegistryImpl {
    constructor() {
        this.typeHierarchy = Object.create(null);
        this.tokenTypeById = Object.create(null);
        this.tokenModifierById = Object.create(null);
        this.currentTypeNumber = 0;
        this.currentModifierBit = 1;
        this.tokenStylingDefaultRules = [];
        this.tokenStylingSchema = {
            type: 'object',
            properties: {},
            patternProperties: {
                [semantic_tokens_registry_1.selectorPattern]: (0, semantic_tokens_registry_1.getStylingSchemeEntry)(),
            },
            additionalProperties: false,
            definitions: {
                style: {
                    type: 'object',
                    description: (0, localize_1.localize)('schema.token.settings', 'Colors and styles for the token.'),
                    properties: {
                        foreground: {
                            type: 'string',
                            description: (0, localize_1.localize)('schema.token.foreground', 'Foreground color for the token.'),
                            format: 'color-hex',
                            default: '#ff0000',
                        },
                        background: {
                            type: 'string',
                            deprecationMessage: (0, localize_1.localize)('schema.token.background.warning', 'Token background colors are currently not supported.'),
                        },
                        fontStyle: {
                            type: 'string',
                            description: (0, localize_1.localize)('schema.token.fontStyle', "Sets the all font styles of the rule: 'italic', 'bold' or 'underline' or a combination. All styles that are not listed are unset. The empty string unsets all styles."),
                            pattern: semantic_tokens_registry_1.fontStylePattern,
                            patternErrorMessage: (0, localize_1.localize)('schema.fontStyle.error', "Font style must be 'italic', 'bold' or 'underline' or a combination. The empty string unsets all styles."),
                            defaultSnippets: [
                                {
                                    label: (0, localize_1.localize)('schema.token.fontStyle.none', 'None (clear inherited style)'),
                                    bodyText: '""',
                                },
                                { body: 'italic' },
                                { body: 'bold' },
                                { body: 'underline' },
                                { body: 'italic underline' },
                                { body: 'bold underline' },
                                { body: 'italic bold underline' },
                            ],
                        },
                        bold: {
                            type: 'boolean',
                            description: (0, localize_1.localize)('schema.token.bold', "Sets or unsets the font style to bold. Note, the presence of 'fontStyle' overrides this setting."),
                        },
                        italic: {
                            type: 'boolean',
                            description: (0, localize_1.localize)('schema.token.italic', "Sets or unsets the font style to italic. Note, the presence of 'fontStyle' overrides this setting."),
                        },
                        underline: {
                            type: 'boolean',
                            description: (0, localize_1.localize)('schema.token.underline', "Sets or unsets the font style to underline. Note, the presence of 'fontStyle' overrides this setting."),
                        },
                    },
                    defaultSnippets: [{ body: { foreground: '${1:#FF0000}', fontStyle: '${2:bold}' } }],
                },
            },
        };
    }
    parseTokenSelector(selectorString, language) {
        const selector = (0, semantic_tokens_registry_1.parseClassifierString)(selectorString, language);
        if (!selector.type) {
            return {
                match: () => -1,
                id: '$invalid',
            };
        }
        return {
            match: (type, modifiers, language) => {
                let score = 0;
                if (selector.language !== undefined) {
                    if (selector.language !== language) {
                        return -1;
                    }
                    score += 10;
                }
                if (selector.type !== semantic_tokens_registry_1.TOKEN_TYPE_WILDCARD) {
                    const hierarchy = this.getTypeHierarchy(type);
                    const level = hierarchy.indexOf(selector.type);
                    if (level === -1) {
                        return -1;
                    }
                    score += 100 - level;
                }
                // all selector modifiers must be present
                for (const selectorModifier of selector.modifiers) {
                    if (modifiers.indexOf(selectorModifier) === -1) {
                        return -1;
                    }
                }
                return score + selector.modifiers.length * 100;
            },
            id: `${[selector.type, ...selector.modifiers.sort()].join('.')}${selector.language !== undefined ? ':' + selector.language : ''}`,
        };
    }
    registerTokenStyleDefault(selector, defaults) {
        this.tokenStylingDefaultRules.push({ selector, defaults });
    }
    deregisterTokenStyleDefault(selector) {
        const selectorString = selector.id;
        this.tokenStylingDefaultRules = this.tokenStylingDefaultRules.filter((r) => r.selector.id !== selectorString);
    }
    getTokenStylingDefaultRules() {
        return this.tokenStylingDefaultRules;
    }
    registerTokenModifier(id, description, deprecationMessage) {
        if (!id.match(semantic_tokens_registry_1.typeAndModifierIdPattern)) {
            throw new Error('Invalid token modifier id.');
        }
        const num = this.currentModifierBit;
        this.currentModifierBit = this.currentModifierBit * 2;
        const tokenStyleContribution = {
            num,
            id,
            description,
            deprecationMessage,
        };
        this.tokenModifierById[id] = tokenStyleContribution;
        this.tokenStylingSchema.properties[`*.${id}`] = (0, semantic_tokens_registry_1.getStylingSchemeEntry)(description, deprecationMessage);
    }
    registerTokenType(id, description, superType, deprecationMessage) {
        if (!id.match(semantic_tokens_registry_1.typeAndModifierIdPattern)) {
            throw new Error('Invalid token type id.');
        }
        if (superType && !superType.match(semantic_tokens_registry_1.typeAndModifierIdPattern)) {
            throw new Error('Invalid token super type id.');
        }
        const num = this.currentTypeNumber++;
        const tokenStyleContribution = {
            num,
            id,
            superType,
            description,
            deprecationMessage,
        };
        this.tokenTypeById[id] = tokenStyleContribution;
        const stylingSchemeEntry = (0, semantic_tokens_registry_1.getStylingSchemeEntry)(description, deprecationMessage);
        this.tokenStylingSchema.properties[id] = stylingSchemeEntry;
        this.typeHierarchy = Object.create(null);
    }
    getTypeHierarchy(typeId) {
        let hierarchy = this.typeHierarchy[typeId];
        if (!hierarchy) {
            this.typeHierarchy[typeId] = hierarchy = [typeId];
            let type = this.tokenTypeById[typeId];
            while (type && type.superType) {
                hierarchy.push(type.superType);
                type = this.tokenTypeById[type.superType];
            }
        }
        return hierarchy;
    }
};
SemanticTokenRegistryImpl = tslib_1.__decorate([
    (0, di_1.Injectable)()
], SemanticTokenRegistryImpl);
exports.SemanticTokenRegistryImpl = SemanticTokenRegistryImpl;
//# sourceMappingURL=semantic-tokens-registry.js.map