"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ThemeData = void 0;
const tslib_1 = require("tslib");
const di_1 = require("@opensumi/di");
const ide_core_browser_1 = require("@opensumi/ide-core-browser");
const common_1 = require("@opensumi/ide-file-service/lib/common");
const color_1 = require("../common/color");
const editor_1 = require("../common/color-tokens/editor");
const plistParser_1 = require("../common/plistParser");
const semantic_tokens_registry_1 = require("../common/semantic-tokens-registry");
const theme_service_1 = require("../common/theme.service");
const themeCompatibility_1 = require("../common/themeCompatibility");
const default_theme_1 = tslib_1.__importDefault(require("./default-theme"));
function getScopeMatcher(rule) {
    const ruleScope = rule.scope;
    if (!ruleScope || !rule.settings) {
        return semantic_tokens_registry_1.noMatch;
    }
    const matchers = [];
    if (Array.isArray(ruleScope)) {
        for (const rs of ruleScope) {
            (0, semantic_tokens_registry_1.createMatchers)(rs, semantic_tokens_registry_1.nameMatcher, matchers);
        }
    }
    else {
        (0, semantic_tokens_registry_1.createMatchers)(ruleScope, semantic_tokens_registry_1.nameMatcher, matchers);
    }
    if (matchers.length === 0) {
        return semantic_tokens_registry_1.noMatch;
    }
    return (scope) => {
        let max = matchers[0].matcher(scope);
        for (let i = 1; i < matchers.length; i++) {
            max = Math.max(max, matchers[i].matcher(scope));
        }
        return max;
    };
}
function isSemanticTokenColorizationSetting(style) {
    return (style &&
        ((0, ide_core_browser_1.isString)(style.foreground) ||
            (0, ide_core_browser_1.isString)(style.fontStyle) ||
            (0, ide_core_browser_1.isBoolean)(style.italic) ||
            (0, ide_core_browser_1.isBoolean)(style.underline) ||
            (0, ide_core_browser_1.isBoolean)(style.bold)));
}
let ThemeData = class ThemeData {
    constructor() {
        this.themeSettings = [];
        this.colors = {};
        this.encodedTokensColors = [];
        this.rules = [];
        this.base = 'vs-dark';
        this.inherit = false;
        this.colorMap = {};
        this.hasDefaultTokens = false;
        this.customSettings = [];
        this.tokenColorIndex = undefined; // created on demand
        this.semanticTokenRules = [];
    }
    initializeFromData(data) {
        this.id = data.id;
        this.name = data.name;
        this.colors = data.colors;
        this.encodedTokensColors = data.encodedTokensColors;
        this.themeSettings = data.themeSettings;
        this.rules = data.rules;
        this.base = data.base;
        this.inherit = data.inherit;
    }
    get type() {
        switch (this.base) {
            case theme_service_1.VS_LIGHT_THEME_NAME:
                return theme_service_1.ColorScheme.LIGHT;
            case theme_service_1.HC_BLACK_THEME_NAME:
                return theme_service_1.ColorScheme.HIGH_CONTRAST_DARK;
            case theme_service_1.HC_LIGHT_THEME_NAME:
                return theme_service_1.ColorScheme.HIGH_CONTRAST_LIGHT;
            default:
                return theme_service_1.ColorScheme.DARK;
        }
    }
    getDefaultTheme() {
        return default_theme_1.default;
    }
    async initializeThemeData(id, name, base, themeLocation) {
        this.id = id;
        this.name = name;
        this.base = base;
        await this.loadColorTheme(themeLocation, this.themeSettings, this.colorMap);
        // eslint-disable-next-line guard-for-in
        for (const key in this.colorMap) {
            this.colors[key] = color_1.Color.Format.CSS.formatHexA(this.colorMap[key]);
        }
    }
    loadCustomTokens(customSettings) {
        this.rules = [];
        // const affectedScopes: string[] = customSettings.map((setting) => setting.scope).filter((t) => !!t);
        this.doInitTokenRules();
        for (const setting of customSettings) {
            this.transform(setting, (rule) => {
                const existIndex = this.rules.findIndex((item) => item.token === rule.token);
                if (existIndex > -1) {
                    this.rules.splice(existIndex, 1, rule);
                }
                else {
                    this.rules.push(rule);
                }
            });
        }
        this.customSettings = customSettings;
    }
    get settings() {
        return this.themeSettings.concat(this.customSettings);
    }
    get tokenColors() {
        const result = [];
        const foreground = this.colorMap[editor_1.editorForeground];
        const background = this.colorMap[editor_1.editorBackground];
        result.push({
            settings: {
                foreground: normalizeColor(foreground),
                background: normalizeColor(background),
            },
        });
        let hasDefaultTokens = false;
        function addRule(rule) {
            if (rule.scope && rule.settings) {
                if (rule.scope === 'token.info-token') {
                    hasDefaultTokens = true;
                }
                result.push({
                    scope: rule.scope,
                    settings: {
                        foreground: normalizeColor(rule.settings.foreground),
                        background: normalizeColor(rule.settings.background),
                        fontStyle: rule.settings.fontStyle,
                    },
                });
            }
        }
        this.settings.map(addRule);
        if (!hasDefaultTokens) {
            defaultThemeColors[this.type].forEach(addRule);
        }
        return result;
    }
    // transform settings & init rules
    doInitTokenRules() {
        for (const setting of this.themeSettings) {
            this.transform(setting, (rule) => this.rules.push(rule));
        }
        if (!this.hasDefaultTokens) {
            defaultThemeColors[(0, theme_service_1.getThemeType)(this.base)].forEach((setting) => {
                this.transform(setting, (rule) => this.rules.push(rule));
            });
        }
    }
    safeParseJSON(content) {
        let json;
        try {
            json = (0, ide_core_browser_1.parseWithComments)(content);
            return json;
        }
        catch (error) {
            return this.logger.error('Theme data parse error.', content);
        }
    }
    readSemanticTokenRule(selectorString, settings) {
        const selector = this.semanticTokenRegistry.parseTokenSelector(selectorString);
        let style;
        if (typeof settings === 'string') {
            style = semantic_tokens_registry_1.TokenStyle.fromSettings(settings, undefined);
        }
        else if (isSemanticTokenColorizationSetting(settings)) {
            style = semantic_tokens_registry_1.TokenStyle.fromSettings(settings.foreground, settings.fontStyle, settings.bold, settings.underline, settings.italic);
        }
        if (style) {
            return { selector, style };
        }
        return undefined;
    }
    async loadColorTheme(themeLocation, resultRules, resultColors) {
        const timer = this.reporter.time(ide_core_browser_1.REPORT_NAME.THEME_LOAD);
        const ret = await this.fileServiceClient.readFile(themeLocation.toString());
        const themeContent = ret.content.toString();
        timer.timeEnd(themeLocation.toString());
        const themeLocationPath = themeLocation.path.toString();
        if (/\.json$/.test(themeLocationPath)) {
            const theme = this.safeParseJSON(themeContent);
            let includeCompletes = Promise.resolve(null);
            if (theme.include) {
                this.inherit = true;
                // http 的不作支持
                const includePath = themeLocation.path.dir.join(theme.include.replace(/^\.\//, ''));
                const includeLocation = themeLocation.withPath(includePath);
                includeCompletes = this.loadColorTheme(includeLocation, resultRules, resultColors);
            }
            await includeCompletes;
            // settings
            if (Array.isArray(theme.settings)) {
                (0, themeCompatibility_1.convertSettings)(theme.settings, resultRules, resultColors);
                return null;
            }
            // semanticHighlighting enable/disabled
            this.semanticHighlighting = theme.semanticHighlighting;
            // semanticTokenColors
            const semanticTokenColors = theme.semanticTokenColors;
            if (semanticTokenColors && typeof semanticTokenColors === 'object') {
                // eslint-disable-next-line guard-for-in
                for (const key in semanticTokenColors) {
                    try {
                        const rule = this.readSemanticTokenRule(key, semanticTokenColors[key]);
                        if (rule) {
                            this.semanticTokenRules.push(rule);
                        }
                    }
                    catch (err) {
                        // ignore error
                    }
                }
            }
            // colors
            const colors = theme.colors;
            if (colors) {
                if (typeof colors !== 'object') {
                    return Promise.reject(new Error((0, ide_core_browser_1.localize)('error.invalidformat.colors', "Problem parsing color theme file: {0}. Property 'colors' is not of type 'object'.", themeLocation.toString())));
                }
                // new JSON color themes format
                // eslint-disable-next-line guard-for-in
                for (const colorId in colors) {
                    const colorHex = colors[colorId];
                    if (typeof colorHex === 'string') {
                        // ignore colors tht are null
                        resultColors[colorId] = color_1.Color.fromHex(colors[colorId]);
                    }
                }
            }
            // tokenColors
            const tokenColors = theme.tokenColors;
            if (tokenColors) {
                if (Array.isArray(tokenColors)) {
                    resultRules.push(...tokenColors);
                    return null;
                }
                else if (typeof tokenColors === 'string') {
                    const tokenPath = themeLocation.path.dir.join(tokenColors.replace(/^\.\//, ''));
                    const tokenLocation = themeLocation.withPath(tokenPath);
                    // tmTheme
                    return this.loadSyntaxTokens(tokenLocation);
                }
                else {
                    return Promise.reject(new Error((0, ide_core_browser_1.localize)('error.invalidformat.tokenColors', "Problem parsing color theme file: {0}. Property 'tokenColors' should be either an array specifying colors or a path to a TextMate theme file", themeLocation.toString())));
                }
            }
            return null;
        }
        else {
            return this.loadSyntaxTokens(themeLocation);
        }
    }
    async loadSyntaxTokens(themeLocation) {
        const ret = await this.fileServiceClient.readFile(themeLocation.toString());
        try {
            const theme = (0, plistParser_1.parse)(ret.content.toString());
            const settings = theme.settings;
            if (!Array.isArray(settings)) {
                return Promise.reject(new Error((0, ide_core_browser_1.localize)('error.plist.invalidformat', "Problem parsing tmTheme file: {0}. 'settings' is not array.")));
            }
            (0, themeCompatibility_1.convertSettings)(settings, this.themeSettings, this.colorMap);
            return Promise.resolve(settings);
        }
        catch (e) {
            return Promise.reject(new Error((0, ide_core_browser_1.localize)('error.cannotparse', 'Problems parsing tmTheme file: {0}', e.message)));
        }
    }
    // 将 ITokenColorizationRule 转化为 ITokenThemeRule
    transform(tokenColor, acceptor) {
        if (tokenColor.scope && tokenColor.settings && tokenColor.scope === 'token.info-token') {
            this.hasDefaultTokens = true;
        }
        if (typeof tokenColor.scope === 'undefined') {
            tokenColor.scope = [''];
        }
        else if (typeof tokenColor.scope === 'string') {
            // tokenColor.scope = tokenColor.scope.split(',').map((scope: string) => scope.trim()); // ?
            tokenColor.scope = [tokenColor.scope];
        }
        for (const scope of tokenColor.scope) {
            // Converting numbers into a format that monaco understands
            const settings = Object.keys(tokenColor.settings).reduce((previous, current) => {
                let value = tokenColor.settings[current];
                if (current !== 'foreground' && current !== 'background' && current !== 'fontStyle') {
                    delete tokenColor.settings[current];
                    return previous;
                }
                if (current !== 'fontStyle' && typeof value === 'string') {
                    if (value.indexOf('#') === -1) {
                        // 兼容 white、red 类型色值
                        const color = color_1.Color[value];
                        if (color) {
                            value = color_1.Color.Format.CSS.formatHex(color);
                            tokenColor.settings[current] = value;
                        }
                        else {
                            // 去掉主题瞎写的值
                            delete tokenColor.settings[current];
                            return previous;
                        }
                    }
                    else {
                        const color = color_1.Color.fromHex(value);
                        value = color_1.Color.Format.CSS.formatHex(color);
                        // 主题只会识别 Hex 的色值
                        tokenColor.settings[current] = value;
                    }
                }
                previous[current] = value;
                return previous;
            }, {});
            acceptor(Object.assign(Object.assign({}, settings), { token: scope }));
        }
    }
    getTokenColorIndex() {
        // collect all colors that tokens can have
        if (!this.tokenColorIndex) {
            const index = new TokenColorIndex();
            for (const color of this.encodedTokensColors) {
                index.add(color);
            }
            this.tokenColors.forEach((rule) => {
                index.add(rule.settings.foreground);
                index.add(rule.settings.background);
            });
            this.semanticTokenRules.forEach((r) => index.add(r.style.foreground));
            this.semanticTokenRegistry.getTokenStylingDefaultRules().forEach((r) => {
                const defaultColor = r.defaults[this.type];
                if (defaultColor && typeof defaultColor === 'object') {
                    index.add(defaultColor.foreground);
                }
            });
            this.tokenColorIndex = index;
        }
        return this.tokenColorIndex;
    }
    getTokenStyle(type, modifiers, language, useDefault = true, definitions = {}) {
        const result = {
            foreground: undefined,
            bold: undefined,
            underline: undefined,
            italic: undefined,
        };
        const score = {
            foreground: -1,
            bold: -1,
            underline: -1,
            italic: -1,
        };
        let hasUndefinedStyleProperty = false;
        // eslint-disable-next-line guard-for-in
        for (const k in score) {
            const key = k;
            if (score[key] === -1) {
                hasUndefinedStyleProperty = true;
            }
            else {
                score[key] = Number.MAX_VALUE; // set it to the max, so it won't be replaced by a default
            }
        }
        function _processStyle(matchScore, style, definition) {
            if (style.foreground && score.foreground <= matchScore) {
                score.foreground = matchScore;
                result.foreground = style.foreground;
                definitions.foreground = definition;
            }
            for (const p of ['bold', 'underline', 'italic']) {
                const property = p;
                const info = style[property];
                if (info !== undefined) {
                    if (score[property] <= matchScore) {
                        score[property] = matchScore;
                        result[property] = info;
                        definitions[property] = definition;
                    }
                }
            }
        }
        function _processSemanticTokenRule(rule) {
            const matchScore = rule.selector.match(type, modifiers, language);
            if (matchScore >= 0) {
                _processStyle(matchScore, rule.style, rule);
            }
        }
        this.semanticTokenRules.forEach(_processSemanticTokenRule);
        if (hasUndefinedStyleProperty) {
            for (const rule of this.semanticTokenRegistry.getTokenStylingDefaultRules()) {
                const matchScore = rule.selector.match(type, modifiers, language);
                if (matchScore >= 0) {
                    let style;
                    if (rule.defaults.scopesToProbe) {
                        style = this.resolveScopes(rule.defaults.scopesToProbe);
                        if (style) {
                            _processStyle(matchScore, style, rule.defaults.scopesToProbe);
                        }
                    }
                    if (!style && useDefault !== false) {
                        const tokenStyleValue = rule.defaults[this.type];
                        style = this.resolveTokenStyleValue(tokenStyleValue);
                        if (style) {
                            _processStyle(matchScore, style, tokenStyleValue);
                        }
                    }
                }
            }
        }
        return semantic_tokens_registry_1.TokenStyle.fromData(result);
    }
    /**
     * @param tokenStyleValue Resolve a tokenStyleValue in the context of a theme
     */
    resolveTokenStyleValue(tokenStyleValue) {
        if (tokenStyleValue === undefined) {
            return undefined;
        }
        else if (typeof tokenStyleValue === 'string') {
            const { type, modifiers, language } = (0, semantic_tokens_registry_1.parseClassifierString)(tokenStyleValue, '');
            return this.getTokenStyle(type, modifiers, language);
        }
        else if (typeof tokenStyleValue === 'object') {
            return tokenStyleValue;
        }
        return undefined;
    }
    resolveScopes(scopes, definitions) {
        if (!this.themeTokenScopeMatchers) {
            this.themeTokenScopeMatchers = this.themeSettings.map(getScopeMatcher);
        }
        for (const scope of scopes) {
            let foreground;
            let fontStyle;
            let foregroundScore = -1;
            let fontStyleScore = -1;
            let fontStyleThemingRule;
            let foregroundThemingRule;
            function findTokenStyleForScopeInScopes(scopeMatchers, themingRules) {
                for (let i = 0; i < scopeMatchers.length; i++) {
                    const score = scopeMatchers[i](scope);
                    if (score >= 0) {
                        const themingRule = themingRules[i];
                        const settings = themingRules[i].settings;
                        if (score >= foregroundScore && settings.foreground) {
                            foreground = settings.foreground;
                            foregroundScore = score;
                            foregroundThemingRule = themingRule;
                        }
                        if (score >= fontStyleScore && (0, ide_core_browser_1.isString)(settings.fontStyle)) {
                            fontStyle = settings.fontStyle;
                            fontStyleScore = score;
                            fontStyleThemingRule = themingRule;
                        }
                    }
                }
            }
            findTokenStyleForScopeInScopes(this.themeTokenScopeMatchers, this.themeSettings);
            if (foreground !== undefined || fontStyle !== undefined) {
                if (definitions) {
                    definitions.foreground = foregroundThemingRule;
                    definitions.bold = definitions.italic = definitions.underline = fontStyleThemingRule;
                    definitions.scope = scope;
                }
                return semantic_tokens_registry_1.TokenStyle.fromSettings(foreground, fontStyle);
            }
        }
        return undefined;
    }
};
tslib_1.__decorate([
    (0, di_1.Autowired)(common_1.IFileServiceClient),
    tslib_1.__metadata("design:type", Object)
], ThemeData.prototype, "fileServiceClient", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(ide_core_browser_1.ILogger),
    tslib_1.__metadata("design:type", Object)
], ThemeData.prototype, "logger", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(ide_core_browser_1.IReporterService),
    tslib_1.__metadata("design:type", Object)
], ThemeData.prototype, "reporter", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(semantic_tokens_registry_1.ISemanticTokenRegistry),
    tslib_1.__metadata("design:type", Object)
], ThemeData.prototype, "semanticTokenRegistry", void 0);
ThemeData = tslib_1.__decorate([
    (0, di_1.Injectable)({ multiple: true })
], ThemeData);
exports.ThemeData = ThemeData;
function normalizeColor(color) {
    if (!color) {
        return undefined;
    }
    if (typeof color !== 'string') {
        color = color_1.Color.Format.CSS.formatHexA(color, true);
    }
    const len = color.length;
    if (color.charCodeAt(0) !== 35 /* CharCode.Hash */ || (len !== 4 && len !== 5 && len !== 7 && len !== 9)) {
        return undefined;
    }
    const result = [35 /* CharCode.Hash */];
    for (let i = 1; i < len; i++) {
        const upper = hexUpper(color.charCodeAt(i));
        if (!upper) {
            return undefined;
        }
        result.push(upper);
        if (len === 4 || len === 5) {
            result.push(upper);
        }
    }
    if (result.length === 9 && result[7] === 70 /* CharCode.F */ && result[8] === 70 /* CharCode.F */) {
        result.length = 7;
    }
    return String.fromCharCode(...result);
}
function hexUpper(charCode) {
    if ((charCode >= 48 /* CharCode.Digit0 */ && charCode <= 57 /* CharCode.Digit9 */) ||
        (charCode >= 65 /* CharCode.A */ && charCode <= 70 /* CharCode.F */)) {
        return charCode;
    }
    else if (charCode >= 97 /* CharCode.a */ && charCode <= 102 /* CharCode.f */) {
        return charCode - 97 /* CharCode.a */ + 65 /* CharCode.A */;
    }
    return 0;
}
class TokenColorIndex {
    constructor() {
        this._lastColorId = 0;
        this._id2color = [];
        this._color2id = Object.create(null);
    }
    add(color) {
        color = normalizeColor(color);
        if (color === undefined) {
            return 0;
        }
        let value = this._color2id[color];
        if (value) {
            return value;
        }
        value = ++this._lastColorId;
        this._color2id[color] = value;
        this._id2color[value] = color;
        return value;
    }
    get(color) {
        color = normalizeColor(color);
        if (color === undefined) {
            return 0;
        }
        const value = this._color2id[color];
        if (value) {
            return value;
        }
        return 0;
    }
    asArray() {
        return this._id2color.slice(0);
    }
}
const defaultThemeColors = {
    light: [
        { scope: 'token.info-token', settings: { foreground: '#316bcd' } },
        { scope: 'token.warn-token', settings: { foreground: '#cd9731' } },
        { scope: 'token.error-token', settings: { foreground: '#cd3131' } },
        { scope: 'token.debug-token', settings: { foreground: '#800080' } },
    ],
    dark: [
        { scope: 'token.info-token', settings: { foreground: '#6796e6' } },
        { scope: 'token.warn-token', settings: { foreground: '#cd9731' } },
        { scope: 'token.error-token', settings: { foreground: '#f44747' } },
        { scope: 'token.debug-token', settings: { foreground: '#b267e6' } },
    ],
    hcLight: [
        { scope: 'token.info-token', settings: { foreground: '#316bcd' } },
        { scope: 'token.warn-token', settings: { foreground: '#cd9731' } },
        { scope: 'token.error-token', settings: { foreground: '#cd3131' } },
        { scope: 'token.debug-token', settings: { foreground: '#800080' } },
    ],
    hcDark: [
        { scope: 'token.info-token', settings: { foreground: '#6796e6' } },
        { scope: 'token.warn-token', settings: { foreground: '#008000' } },
        { scope: 'token.error-token', settings: { foreground: '#FF0000' } },
        { scope: 'token.debug-token', settings: { foreground: '#b267e6' } },
    ],
};
//# sourceMappingURL=theme-data.js.map