"use strict";
/* ---------------------------------------------------------------------------------------------
 *  Copyright (c) Microsoft Corporation. All rights reserved.
 *  Licensed under the MIT License. See License.txt in the project root for license information.
 *--------------------------------------------------------------------------------------------*/
Object.defineProperty(exports, "__esModule", { value: true });
exports.getIconRegistry = exports.registerIcon = exports.Extensions = void 0;
// Some code copied and modified from https://github.com/microsoft/vscode/blob/1.86.0/src/vs/platform/theme/common/iconRegistry.ts
const ide_core_common_1 = require("@opensumi/ide-core-common");
const codicons_1 = require("@opensumi/ide-core-common/lib/codicons");
const icons_1 = require("./icons");
exports.Extensions = {
    IconContribution: 'base.contributions.icons',
};
class IconRegistry {
    constructor() {
        this._onDidChange = new ide_core_common_1.Emitter();
        this.onDidChange = this._onDidChange.event;
        this.iconSchema = {
            definitions: {
                icons: {
                    type: 'object',
                    properties: {
                        fontId: {
                            type: 'string',
                            description: (0, ide_core_common_1.localize)('iconDefinition.fontId', 'The id of the font to use. If not set, the font that is defined first is used.'),
                        },
                        fontCharacter: {
                            type: 'string',
                            description: (0, ide_core_common_1.localize)('iconDefinition.fontCharacter', 'The font character associated with the icon definition.'),
                        },
                    },
                    additionalProperties: false,
                    defaultSnippets: [{ body: { fontCharacter: '\\\\e030' } }],
                },
            },
            type: 'object',
            properties: {},
        };
        this.iconReferenceSchema = {
            type: 'string',
            pattern: `^${ide_core_common_1.ThemeIcon.iconNameExpression}$`,
            enum: [],
            enumDescriptions: [],
        };
        this.iconsById = {};
        this.iconFontsById = {};
        this.sumiIconsById = {};
    }
    registerIcon(id, defaults, description, deprecationMessage) {
        const existing = this.iconsById[id];
        if (existing) {
            if (description && !existing.description) {
                existing.description = description;
                this.iconSchema.properties[id].markdownDescription = `${description} $(${id})`;
                const enumIndex = this.iconReferenceSchema.enum.indexOf(id);
                if (enumIndex !== -1) {
                    this.iconReferenceSchema.enumDescriptions[enumIndex] = description;
                }
                this._onDidChange.fire();
            }
            return existing;
        }
        const iconContribution = {
            id,
            defaults,
            description,
            deprecationMessage,
        };
        this.iconsById[id] = iconContribution;
        const propertySchema = { $ref: '#/definitions/icons' };
        if (deprecationMessage) {
            propertySchema.deprecationMessage = deprecationMessage;
        }
        if (description) {
            propertySchema.markdownDescription = `${description}: $(${id})`;
        }
        this.iconSchema.properties[id] = propertySchema;
        this.iconReferenceSchema.enum.push(id);
        this.iconReferenceSchema.enumDescriptions.push(description || '');
        this._onDidChange.fire();
        return { id };
    }
    registerSumiIcon(id, defaults, description, deprecationMessage) {
        const existing = this.sumiIconsById[id];
        if (existing) {
            if (description && !existing.description) {
                existing.description = description;
                this.iconSchema.properties[id].markdownDescription = `${description} $(${id})`;
                const enumIndex = this.iconReferenceSchema.enum.indexOf(id);
                if (enumIndex !== -1) {
                    this.iconReferenceSchema.enumDescriptions[enumIndex] = description;
                }
                this._onDidChange.fire();
            }
            return existing;
        }
        const iconContribution = {
            id,
            defaults,
            description,
            deprecationMessage,
        };
        this.sumiIconsById[id] = iconContribution;
        const propertySchema = { $ref: '#/definitions/icons' };
        if (deprecationMessage) {
            propertySchema.deprecationMessage = deprecationMessage;
        }
        if (description) {
            propertySchema.markdownDescription = `${description}: $(${id})`;
        }
        this.iconSchema.properties[id] = propertySchema;
        this.iconReferenceSchema.enum.push(id);
        this.iconReferenceSchema.enumDescriptions.push(description || '');
        this._onDidChange.fire();
        return { id };
    }
    deregisterIcon(id) {
        delete this.iconsById[id];
        delete this.iconSchema.properties[id];
        const index = this.iconReferenceSchema.enum.indexOf(id);
        if (index !== -1) {
            this.iconReferenceSchema.enum.splice(index, 1);
            this.iconReferenceSchema.enumDescriptions.splice(index, 1);
        }
        this._onDidChange.fire();
    }
    getIcons(isSumi) {
        if (isSumi) {
            return Object.keys(this.sumiIconsById).map((id) => this.sumiIconsById[id]);
        }
        return Object.keys(this.iconsById).map((id) => this.iconsById[id]);
    }
    getIcon(id) {
        return this.iconsById[id];
    }
    getSumiIcon(id) {
        return this.sumiIconsById[id];
    }
    getIconSchema() {
        return this.iconSchema;
    }
    getIconReferenceSchema() {
        return this.iconReferenceSchema;
    }
    registerIconFont(id, definition) {
        const existing = this.iconFontsById[id];
        if (existing) {
            return existing;
        }
        this.iconFontsById[id] = definition;
        this._onDidChange.fire();
        return definition;
    }
    deregisterIconFont(id) {
        delete this.iconFontsById[id];
    }
    getIconFont(id) {
        return this.iconFontsById[id];
    }
    toString() {
        const sorter = (i1, i2) => i1.id.localeCompare(i2.id);
        const classNames = (i) => {
            while (ide_core_common_1.ThemeIcon.isThemeIcon(i.defaults)) {
                i = this.iconsById[i.defaults.id];
            }
            return `codicon codicon-${i ? i.id : ''}`;
        };
        const reference = [];
        reference.push('| preview     | identifier                        | default codicon ID                | description');
        reference.push('| ----------- | --------------------------------- | --------------------------------- | --------------------------------- |');
        const contributions = Object.keys(this.iconsById).map((key) => this.iconsById[key]);
        for (const i of contributions.filter((i) => !!i.description).sort(sorter)) {
            reference.push(`|<i class="${classNames(i)}"></i>|${i.id}|${ide_core_common_1.ThemeIcon.isThemeIcon(i.defaults) ? i.defaults.id : i.id}|${i.description || ''}|`);
        }
        reference.push('| preview     | identifier                        ');
        reference.push('| ----------- | --------------------------------- |');
        for (const i of contributions.filter((i) => !ide_core_common_1.ThemeIcon.isThemeIcon(i.defaults)).sort(sorter)) {
            reference.push(`|<i class="${classNames(i)}"></i>|${i.id}|`);
        }
        return reference.join('\n');
    }
}
const iconRegistry = new IconRegistry();
function registerIcon(id, defaults, description, deprecationMessage) {
    return iconRegistry.registerIcon(id, defaults, description, deprecationMessage);
}
exports.registerIcon = registerIcon;
function getIconRegistry() {
    return iconRegistry;
}
exports.getIconRegistry = getIconRegistry;
// 初始化注册 vscode 定义 icon
function initialize() {
    const codiconFontCharacters = (0, codicons_1.getCodiconFontCharacters)();
    for (const icon in codiconFontCharacters) {
        if (Object.hasOwn(codiconFontCharacters, icon)) {
            const fontCharacter = '\\' + codiconFontCharacters[icon].toString(16);
            iconRegistry.registerIcon(icon, { fontCharacter });
        }
    }
    for (const icon in icons_1.sumiIconIdentifier) {
        if (Object.hasOwn(icons_1.sumiIconIdentifier, icon)) {
            const sumiicon = icons_1.sumiIconIdentifier[icon];
            iconRegistry.registerSumiIcon(icon, sumiicon.defaults, sumiicon.description);
        }
    }
}
initialize();
//# sourceMappingURL=icon-registry.js.map