"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.resolveColorValue = exports.lessProminent = exports.oneOf = exports.transparent = exports.lighten = exports.darken = exports.registerColor = exports.getColorRegistry = exports.Extensions = void 0;
const ide_core_common_1 = require("@opensumi/ide-core-common");
const color_1 = require("./color");
// color registry
exports.Extensions = {
    ColorContribution: 'base.contributions.colors',
};
class ColorRegistry {
    constructor() {
        this.onDidColorChanged = new ide_core_common_1.Emitter();
        this.onDidColorChangedEvent = ide_core_common_1.Event.debounce(this.onDidColorChanged.event, () => { }, 500);
        this.colorsById = {};
    }
    registerColor(id, defaults, description, needsTransparency = false, deprecationMessage) {
        const colorContribution = { id, description, defaults, needsTransparency, deprecationMessage };
        this.colorsById[id] = colorContribution;
        this.onDidColorChanged.fire();
        return id;
    }
    deregisterColor(id) {
        delete this.colorsById[id];
    }
    getColors() {
        return Object.keys(this.colorsById).map((id) => this.colorsById[id]);
    }
    resolveDefaultColor(id, theme) {
        const colorDesc = this.colorsById[id];
        if (colorDesc && colorDesc.defaults) {
            const colorValue = colorDesc.defaults[theme.type];
            return resolveColorValue(colorValue, theme);
        }
        return undefined;
    }
    toString() {
        const sorter = (a, b) => {
            const cat1 = a.indexOf('.') === -1 ? 0 : 1;
            const cat2 = b.indexOf('.') === -1 ? 0 : 1;
            if (cat1 !== cat2) {
                return cat1 - cat2;
            }
            return a.localeCompare(b);
        };
        return Object.keys(this.colorsById)
            .sort(sorter)
            .map((k) => `- \`${k}\`: ${this.colorsById[k].description}`)
            .join('\n');
    }
}
const colorRegistry = new ColorRegistry();
function getColorRegistry() {
    return colorRegistry;
}
exports.getColorRegistry = getColorRegistry;
function registerColor(id, defaults, description, needsTransparency, deprecationMessage) {
    return colorRegistry.registerColor(id, defaults, description, needsTransparency, deprecationMessage);
}
exports.registerColor = registerColor;
function darken(colorValue, factor) {
    return (theme) => {
        const color = resolveColorValue(colorValue, theme);
        if (color) {
            return color.darken(factor);
        }
        return undefined;
    };
}
exports.darken = darken;
function lighten(colorValue, factor) {
    return (theme) => {
        const color = resolveColorValue(colorValue, theme);
        if (color) {
            return color.lighten(factor);
        }
        return undefined;
    };
}
exports.lighten = lighten;
function transparent(colorValue, factor) {
    return (theme) => {
        const color = resolveColorValue(colorValue, theme);
        if (color) {
            return color.transparent(factor);
        }
        return undefined;
    };
}
exports.transparent = transparent;
function oneOf(...colorValues) {
    return (theme) => {
        for (const colorValue of colorValues) {
            const color = resolveColorValue(colorValue, theme);
            if (color) {
                return color;
            }
        }
        return undefined;
    };
}
exports.oneOf = oneOf;
function lessProminent(colorValue, backgroundColorValue, factor, transparency) {
    return (theme) => {
        const from = resolveColorValue(colorValue, theme);
        if (from) {
            const backgroundColor = resolveColorValue(backgroundColorValue, theme);
            if (backgroundColor) {
                if (from.isDarkerThan(backgroundColor)) {
                    return color_1.Color.getLighterColor(from, backgroundColor, factor).transparent(transparency);
                }
                return color_1.Color.getDarkerColor(from, backgroundColor, factor).transparent(transparency);
            }
            return from.transparent(factor * transparency);
        }
        return undefined;
    };
}
exports.lessProminent = lessProminent;
/**
 * @param colorValue Resolve a color value in the context of a theme
 */
function resolveColorValue(colorValue, theme) {
    if (colorValue === null) {
        return undefined;
    }
    else if (colorValue === 'transparent') {
        return color_1.Color.fromHex('#00000000');
    }
    else if (typeof colorValue === 'string') {
        if (colorValue[0] === '#') {
            return color_1.Color.fromHex(colorValue);
        }
        return theme.getColor(colorValue);
    }
    else if (colorValue instanceof color_1.Color) {
        return colorValue;
    }
    else if (typeof colorValue === 'function') {
        return colorValue(theme);
    }
    return undefined;
}
exports.resolveColorValue = resolveColorValue;
//# sourceMappingURL=utils.js.map